import socket

from edunet.core.networking.handlers.connection_handler import ConnectionHandler


class WSGIHandler(ConnectionHandler):
    def handle_connection(self, client_socket: socket.socket):
        pass



from abc import ABC, abstractmethod


class ConnectionHandler(ABC):
    """
    Interface for implementing a connection handler used by the TCPListener
    """

    @abstractmethod
    def handle_connection(self, *args, **kwargs):
        """
        This provides the interface that has to be implemented that will be used by the
        TCPListener. The main idea is that a connection handler is provided to the
        TCPListener to handle the socket data as at comes in.

        As an example, you can implement a wsgi handler to handle connections.

        class WSGIHandler(ConnectionHandler):
            def handle_connection(foo, bar):
                pass
        """



class EduRouterException(Exception):
    """
    Base EduRouterException will be used for general application exceptions and to be
    inherited from edunet.by all other exceptions
    """

    pass


class TCPListenerError(EduRouterException):
    pass


class HTTPValidationError(EduRouterException):
    pass


class HTTPDataModelError(EduRouterException):
    pass



import logging
from abc import ABC, abstractmethod

from edunet.core.networking.listeners.listener import Listener

logger = logging.getLogger(__name__)


class Node(ABC):
    """
    Representation of a nodes that will allow to hold an optional service to handle
    requests, or if no service passed will act as a client nodes
    """

    def __init__(self, listener: Listener):
        self.listener = listener

    @abstractmethod
    def start(self, *args, **kwargs) -> None:
        """
        Used to implement how a Node starts
        """

    @abstractmethod
    def stop(self, *args, **kwargs) -> None:
        """
        Used to implement how a Node stops
        """



from abc import abstractmethod, ABC

from edunet.models.base_types import Request, Response


class Application(ABC):
    """
    Base Application implementation for any type of application that simply handles
    request and responses

    class MyApplication(Application):
        def handle_request(self, request):
            # implementation using request object that might be of type HTTP, TCP, UDP
            return some_response
    """

    @abstractmethod
    def handle_request(self, request_data: Request) -> Response:
        """
        Handling the request intended coming from a ConnectionHandler implementation

        e.g.
        class HTTPApplication(Application):
            ...
            def handle_request(self, foo, bar):
                self.custom_logic(foo, bar)
        """



import logging

from edunet.core.applications.application import Application
from edunet.models.http import HTTPRequest, HTTPResponse

logger = logging.getLogger(__name__)


class SimpleHTTPApplication(Application):

    def handle_request(self, request_data: HTTPRequest) -> HTTPResponse:
        """
        Provide a HTTPRequest object to receive a HTTPResponse object.
        """

        logger.info(f"Request data: {request_data}")

        try:
            return HTTPResponse(
                status_code=200,
                status_text="OK",
                body=f"Message received: {request_data}",
            )
        except (SyntaxError, TypeError, ValueError, AttributeError) as e:
            logger.error(f"Could not construct data: {e}")
            return HTTPResponse(
                status_code=500, status_text="Internal Server Error", body=""
            )
        except Exception as e:
            logger.exception(f"Unexpected server error: {e}")
            return HTTPResponse(
                status_code=500, status_text="Internal Server Error", body=""
            )



import logging
import re
from typing import Tuple, Dict

from edunet.exceptions import HTTPValidationError

# Compiled regex patterns for request line, headers, and body
REQUEST_LINE_PATTERN = re.compile(rb"^[A-Z]+ [^\r\n]* HTTP/(1\.[01]|2\.0)$")
HEADER_PATTERN = re.compile(rb"^[^\r\n]+: [^\r\n]*$")
BODY_PATTERN = re.compile(rb".*")

logger = logging.getLogger(__name__)


def validate_and_get_http_request_components(
    request: bytes,
) -> Tuple[bytes, Dict[bytes, bytes], bytes]:
    """
    Validates an HTTP request and returns its parts if valid.

    Args:
    - request (bytes): The HTTP request to validate.

    Returns:
    - tuple: A tuple containing the request line, headers, and body if the request
             is valid, otherwise None.
    """

    # Split the request into components, excluding empty lines
    request_line, headers, body = _get_parts(request)

    # Validate request line
    if not re.match(REQUEST_LINE_PATTERN, request_line):
        logger.error("Invalid request line: %s", request_line)
        raise HTTPValidationError(f"Invalid request line pattern: {request_line}")

    # Validate headers format and parse into key-value pairs
    parsed_headers = {}
    for header in headers:
        if not re.match(HEADER_PATTERN, header):
            logger.error("Invalid header pattern: %s", header)
            raise HTTPValidationError(f"Invalid header pattern: {header}")

        # Split header into key and value
        header_name, header_value = header.split(b": ", 1)
        parsed_headers[header_name] = header_value

    return request_line, parsed_headers, body


def _get_parts(request):
    # Split the request into components using line boundaries
    components = request.split(b"\r\n\r\n", 1)

    # Extract the request line and headers
    request_line_and_headers = components[0].split(b"\r\n")
    request_line = request_line_and_headers[0]
    headers = request_line_and_headers[1:]

    # Initialize the body
    body = b""

    # If there are components beyond the request line and headers
    if len(components) > 1:
        body = components[1]

    return request_line, headers, body



import logging
import socket
import threading

from edunet.core.networking.handlers.connection_handler import ConnectionHandler
from edunet.core.networking.listeners.listener import Listener
from edunet.exceptions import TCPListenerError

logger = logging.getLogger(__name__)


class TCPListener(Listener):
    def __init__(self, hostname: str, port: int, connection_handler: ConnectionHandler):
        self.hostname = hostname
        self.port = port

        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.bind((self.hostname, self.port))
        self._is_listening = False

        self.connection_handler = connection_handler

    @property
    def is_listening(self):
        """
        Read only property to check whether the TCPListener is listening and active.

        This is controlled internally through starting and stopping the service.
        e.g. tcp_listener.start() and tcp_listener.stop()
        """
        return self._is_listening

    def accept_connection(self) -> None:
        """
        This method will be used to accept incoming connections. It will only function
        if the service is running. Ensure the start_service() method is called first.
        An exception will be raised otherwise.
        """

        if not self.is_listening:
            raise RuntimeError("Service is not running. Please start service.")

        while self.is_listening:
            try:
                client_socket, _ = self.server_socket.accept()
                logger.info("Connection established.")

                request_data = client_socket.recv(1024)

                logger.info(f"The socket client: {client_socket}")

                threading.Thread(
                    target=self.handle_request, args=(request_data, client_socket)
                ).start()

                # We still want a safeguarded way to break the loop if we find ourselves
                # in a situation where the service goes down while hanging in this loop
                if not self.is_listening:
                    logger.warning("Detected service interruption. Exiting.")
                    break
            except Exception as e:
                # Handle case where the socket is closed while accepting connections
                if not self.is_listening:
                    logger.warning("Service stopped while accepting connections.")
                    break
                else:
                    logger.error(f"Error handling client socket: {e}")

    def handle_request(self, request_data, client_socket: socket.socket) -> None:
        """
        Provide a RequestData object to receive a ResponseData response.
        """
        try:
            res = self.connection_handler.handle_connection(request_data, client_socket)
            logger.info(f"Data to be sent back: {res}")
            client_socket.sendall(res)
            logger.info("Response sent back to client.")
        except socket.error as e:
            logger.error(f"Could not send data: {e}")
            raise TCPListenerError(e)
        finally:
            logger.info("Closing connection")
            self._close_client_socket(client_socket)

    def start(self) -> None:
        """
        Will start the TCPListener service to listen for incoming connections
        """
        try:
            if not self.is_listening:
                logger.info("Starting service.")
                self.server_socket.listen()
                self._is_listening = True

                logger.info(f"Service listening on {self.hostname}:{self.port}")

                self.accept_connection()
            else:
                logger.info("Service already starting. Nothing to do")
        except Exception as e:
            logger.error(f"Could not start service: {e}")
            raise TCPListenerError(f"Could not start service: {e}")

    def stop(self) -> None:
        """
        Will stop the TCPListener service and will no longer accept connections
        """
        logger.info("Stopping service")
        if self.is_listening:
            try:
                self.server_socket.close()
                self._is_listening = False
            except Exception as e:
                logger.error(f"Error shutting down service: {e}")
                raise TCPListenerError(f"Error shutting down service: {e}")

            logger.info("All threads terminated. Service has been stopped.")
        else:
            logger.warning("Service is already stopped. Nothing to do.")

    def _close_client_socket(self, client_socket) -> None:
        """
        Internal method to close a client socket and check if it has been closed
        """
        try:
            client_socket.close()
        except socket.error as e:
            logger.error(f"Socket error closing client socket: {e}")
        else:
            logger.info("Client socket closed successfully")
        finally:
            if client_socket.fileno() == -1:
                logger.info("Client socket successfully closed")
            else:
                logger.warning("Socket not closed")



import logging
from dataclasses import dataclass
from typing import Optional, Dict, TypeVar, Type

from edunet.exceptions import HTTPValidationError, HTTPDataModelError
from edunet.models.base_types import Request, Response
from edunet.validators.http_validators import (
    validate_and_get_http_request_components,
)

logger = logging.getLogger(__name__)


T = TypeVar("T", bound="HTTPRequest")


@dataclass
class HTTPRequest(Request):
    method: bytes
    uri: bytes
    version: bytes
    headers: Dict[bytes, bytes]
    body: bytes

    @classmethod
    def from_bytes(cls: Type[T], data: bytes) -> T:
        try:
            request_line, headers, body = validate_and_get_http_request_components(data)

            # Parse request line
            method, uri, version = request_line.split(b" ")

            return cls(
                method=method,
                uri=uri,
                version=version,
                headers=headers,
                body=body,
            )
        except (AttributeError, HTTPValidationError) as e:
            logger.error(f"Error creating Request object: {e}")
            raise HTTPDataModelError(f"Error creating Request object: {e}")


@dataclass
class HTTPResponse(Response):
    status_code: int
    status_text: str
    body: str
    http_version: str = "HTTP/1.1"
    content_type: Optional[str] = None
    additional_headers: Optional[dict] = None

    def to_bytes(self) -> bytes:
        try:
            # Construct status line
            status_line = (
                f"{self.http_version} {self.status_code} {self.status_text}\r\n".encode(
                    "utf-8"
                )
            )

            # Construct headers
            headers = f"Content-Length: {len(self.body)}\r\n".encode("utf-8")
            if self.content_type:
                headers += f"Content-Type: {self.content_type}\r\n".encode("utf-8")
            if self.additional_headers:
                for key, value in self.additional_headers.items():
                    headers += f"{key}: {value}\r\n".encode("utf-8")
            headers += b"\r\n"

            # Combine status line, headers, and body
            response = status_line + headers + self.body.encode("utf-8")
        except (AttributeError, TypeError) as e:
            logger.error(f"Error creating response message: {e}")
            raise HTTPDataModelError(f"Error creating response message: {e}")

        return response



import logging
import socket

from edunet.core.applications.application import Application
from edunet.core.networking.handlers.connection_handler import ConnectionHandler
from edunet.models.http import HTTPRequest

logger = logging.getLogger(__name__)


class SimpleHTTPConnectionHandler(ConnectionHandler):
    def __init__(self, application: Application):
        self.application = application

    def handle_connection(self, data: bytes, client_socket: socket.socket) -> bytes:
        return self.application.handle_request(HTTPRequest.from_bytes(data)).to_bytes()



import logging

from edunet.core.nodes.node import Node
from edunet.core.networking.listeners.listener import Listener

logger = logging.getLogger(__name__)


class SimpleHTTPNode(Node):
    def __init__(self, listener: Listener):
        super().__init__(listener)
        self.machine_started = False

    def start(self):
        logger.info("Starting Machine and services")
        if not self.machine_started:
            self.machine_started = True
            logger.info("Machine has been started")
            self.listener.start()
            logger.info("Service has been started")
        else:
            logger.info("Machine already started. Nothing to do")

    def stop(self):
        logger.info("Stopping services and machine")
        if self.machine_started:
            self.listener.stop()
            logger.info("Service stopped")
            self.machine_started = False
            logger.info("Machine stopped")
        else:
            logger.info("Machine already stopped. Nothing to do")



from abc import abstractmethod, ABC


class Listener(ABC):
    """
    This base class provides the means to implement any listener type to handle incoming
    network requests.

    For example: TCP or UDP listening
    """

    @abstractmethod
    def accept_connection(self, *args, **kwargs):
        """
        Used to implement how connections are accepted
        """

    @abstractmethod
    def handle_request(self, *args, **kwargs):
        """
        Used to implement how connections should be handled
        """

    @abstractmethod
    def start(self, *args, **kwargs):
        """
        Used to implement the starting of the listener
        """

    @abstractmethod
    def stop(self, *args, **kwargs):
        """
        Used to implement the stopping of the listener
        """
